# core/__init__.py
#
#

""" basic package for the program. """

__copyright__ = "Copyright 2014 B.H.J Thate"
__version__ = 33

## IMPORTS

from core.log import datefmt
from core.utils import *

import collections
import threading
import readline
import logging
import getpass
import _thread
import socket
import select
import random
import queue
import types
import errno
import fcntl
import uuid
import core
import json
import time
import imp
import cgi

import sys
import os
import re

## ERRORS

class Error(BaseException):
    """ basic Exception used in the program. """
    pass

class NoWorkdirSet(BaseException):
    """ a working directory is needed. """
    pass

class Overload(Error):
    """ method is trying to be overloaded. """
    pass

class AttributeSet(Error):
    """ attribute is already set. """
    pass

class DataNotSet(Error):
    """ missing argument attribute. """
    pass

class NoTarget(Error):
    """ missing argument attribute. """
    pass

class NoJSON(Error):
    """ no JSON could be decoded/encoded. """
    pass

class RemoteDisconnect(Error):
    """ Exception thrown when the connection is disconnected. """
    pass

class NotSet(Error):
    """ attribute is not set. """
    pass

class FileNotFoundError(Error):
    """ file is not found. """
    pass

## BOOT

homedir = os.path.expanduser("~")
histfile = j(homedir, ".project_history")

try: hostname = socket.getfqdn()
except: hostname = "localhost"

port = 10102

## BASE

class Object(dict):

    """ basic Object on which the rest of the program is based. """

    def __getattribute__(zelf, *args, **kwargs):
        name = args[0]
        if name == "parsed": return txt_parse(zelf.txt)
        if name == "url": return zelf.get_url("get")
        if name == "url_show": return zelf.get_url("show")
        if name == "what": return get_clsname(zelf)
        if name == "how": return get_how(zelf)
        return dict.__getattribute__(zelf, *args, **kwargs)

    def __getattr__(zelf, name):
        if name in zelf: return zelf[name]
        if name == "_ready": zelf["_ready"] = threading.Event()
        if name == "_status": zelf["_status"] = Object()
        if name == "_state": zelf["_state"] = Object()
        if name == "_run": zelf["_run"] = Object()
        if name == "_no_save": zelf["_no_save"] = Object()
        if name not in zelf: raise AttributeError(name)
        return zelf[name]

    def __setattr__(zelf, name, value):
        if name in zelf and type(zelf[name]) in [types.BuiltinMethodType, types.BuiltinFunctionType, types.MethodType, types.FunctionType]: raise OverloadError(name)
        return dict.__setitem__(zelf, name, value)

    def __iter__(zelf):
        for key in zelf.keys(): yield key

    ## FLEET

    def announce(zelf, *args, **kwargs):
        """ announce to all the running cores. """
        for bot in kernel.fleet: bot.announce(args[0])

    ## STATE

    def define(zelf, *args, **kwargs):
        """ set a attribute on this object. """
        name = args[0]
        value = args[1]
        zelf[name] = value

    def register(zelf, *args, **kwargs):
        """ register a value in a list on this object. """
        o = Object()
        o.name = args[0]
        o.func = args[1]
        logging.info("register %s" % o.how)
        if o.name not in zelf: zelf[o.name] = []
        zelf[o.name].append(o)

    ## GETTERS

    def get_root(zelf, *args, **kwargs):
        """ root directory of the program. """
        return j(os.path.expanduser("~"), *args)

    def get_cmnd(zelf, *args, **kwargs):
        """ determine the command in the zelf.txt attribute, if present. """
        if "txt" in zelf and zelf.txt:
            val = zelf.txt.split()[0]
            if not val: return
            if "cc" in zelf:
                if val[0] != zelf.cc: return
                else: return val[1:]
            return val

    def get_rest(zelf, *args, **kwargs):
        """ get the rest of the txt attribute. """
        try: cmnd, rest = zelf.txt.split(" ", 1) ; return rest
        except ValueError: return ""

    def get_args(zelf, *args, **kwargs):
        """ get the arguments of the txt attribute. """
        if "txt" in zelf: return zelf.txt.split()

    def get_days(zelf, *args, **kwargs):
        """ get the number of days relative to the object's creation time. """
        t1 = time.time()
        t2 = a_time(zelf.get_timed())
        if t2:
            time_diff = float(t1 - t2)
            return nr_days(time_diff)

    def get_timed(zelf, *args, **kwargs):
        """ retrieve the creation time of an object. """
        val = ""
        if "Date" in zelf: val = short_date(zelf.Date)
        elif "date" in zelf: val = short_date(zelf.date)
        elif "published" in zelf: val = short_date(zelf.published)
        elif "timed" in zelf: val = short_date(zelf.timed)
        elif "added" in zelf: val = short_date(zelf.added)
        return val

    def get_parsed(zelf, *args, **kwargs):
        """ parse the txt attribute. """
        rest = zelf.get_rest()
        return txt_parse(rest)
        
    def get_keys(zelf, *args, **kwargs):
        """ skip the unwanted keys e.g those that start with a "_". """
        for key in zelf.keys():
            k = str(key)
            if k.startswith("_"): continue
            yield key

    def get_obj(zelf, *args, **kwargs):
        """ cloned object, with only the proper keys used. """
        res = Object()
        for key in zelf.get_keys(): res[key] = zelf[key]
        return res

    def get_slice(zelf, *args, **kwargs):
        """ take a slice of the Object. """
        o = Object()
        arguments = args[0]
        for arg in arguments:
            try: o[arg] = zelf[arg]
            except KeyError: continue
        return o

    def get_path(zelf, *args, **kwargs):
        """ timestamp of related filename. """         
        if "path" in zelf._no_save: return zelf._no_save.path

    def get_filetime(zelf, *args, **kwargs):
        """ timestamp of related filename. """         
        if "path" in zelf._no_save: return zelf._no_save.path.split(os.sep)[-1]

    def get_url(zelf, *args, **kwargs):
        """ url of the object's file so that it can be retrieved when API server is running. """
        import core.plugs.api as api
        fn = zelf.get_path().split(kernel.cfg.workdir)[-1]
        if fn: return "http://%s:%s/%s%s" % (hostname, port, args[0], fn)

    ## CHECKERS

    def check_wanted(zelf, *args, **kwargs):
        """ whether an object is desired. """
        want = args[0]
        for key, value in want.items():
           if key == "format": continue
           if key not in zelf: continue
           if value.startswith("-"): continue
           if value not in str(zelf[key]): return False
        return True

    def check_notwanted(zelf, *args, **kwargs):
        """ whether an object is not desired. """
        not_want = args[0]
        for key, value in not_want.items():
           if key == "format": continue
           if key not in zelf: continue
           if value in zelf[key]: return True
        return False

    ## INPUT

    def load(zelf, *args, **kwargs):
        """ load a JSON file into this object. """
        if args: path = args[0]
        else: path = zelf._no_save["path"]
        ondisk = zelf.read(path)
        fromdisk = json.loads(ondisk)
        zelf._no_save.path = path
        if "data" in fromdisk: zelf.update(fromdisk["data"])
        else: zelf.update(fromdisk)
        return zelf

    def read(zelf, *args, **kwargs):
        """ read the JSON file from disk. """
        path = args[0]
        try: f = open(path, "r")
        except IOError as ex:
            if ex.errno == errno.ENOENT: return "{}"
            raise
        res = ""
        for line in f:
            if not line.strip().startswith("#"): res += line
        if not res.strip(): return "{}"
        f.close()
        return res

    ## PERSISTENCE

    def write(zelf, *args, **kwargs):
        """ actual write to disk. """
        path = args[0]
        logging.warn("write %s" % path)
        base, fn = os.path.split(os.path.abspath(os.path.expanduser(path)))
        cdir(base)
        datafile = open(path + ".tmp", 'w')
        fcntl.flock(datafile, fcntl.LOCK_EX | fcntl.LOCK_NB)
        datafile.write(zelf.to_json(*args, **kwargs))
        datafile.write("\n")
        fcntl.flock(datafile, fcntl.LOCK_UN)
        datafile.close()
        os.rename(path + ".tmp", path)
        return zelf

    def to_disk(zelf, *args, **kwargs):
        """ create JSON ready to be saved to disk. """
        path = args[0]
        todisk = Object()
        todisk.create_type = zelf.what
        todisk.version = __version__
        zelf.timed = time.ctime(time.time())
        todisk.data = zelf.get_obj()
        todisk.signature = todisk.make_signature()
        try: result = todisk.to_json(indent=2, ensure_ascii=False, sort_keys=True)
        except TypeError: raise NoJSON()
        return result

    def save(zelf, *args, **kwargs):
        """ save JSON to disk. """
        if not args: t = rtime()
        else: t = args[0]
        root = j(kernel.cfg.workdir, t)
        cdir(kernel.cfg.workdir)
        zelf.sync(root)
        return root

    def sync(zelf, *args, **kwargs):
        """ sync JSON to disk. """
        try: path = args[0] ; zelf._no_save.path = path
        except IndexError: path = zelf._no_save.path
        logging.warn("sync %s" % path)
        d, fn = os.path.split(path)
        cdir(d)
        todisk = zelf.to_disk(path, **kwargs)
        datafile = open(os.path.abspath(path) + ".tmp", 'w')
        fcntl.flock(datafile, fcntl.LOCK_EX | fcntl.LOCK_NB)
        datafile.write(headertxt % (__version__, "%s characters" % len(todisk)))
        datafile.write(todisk)
        datafile.write("\n")
        fcntl.flock(datafile, fcntl.LOCK_UN)
        datafile.close()
        os.rename(path + ".tmp", path)
        return zelf

    ## OUTPUT

    def reply(zelf, *args, **kwargs):
        """ send reply to origin. """
        if "outer" in zelf: zelf.outer.write(str(args[0]) + "\n") ; zelf.outer.flush() ; return
        txt = args[0]
        if "channel" in zelf: zelf._target.say(zelf.channel, txt)
        else: zelf._target._raw(txt)

    def say(zelf, *args, **kwargs):
        """ output text through the _target attribute. """
        zelf._target.say(args[0], args[1])

    def ok(zelf, *args, **kwargs):
        """ signal ok. """
        if "_target" in zelf: zelf.reply("ok %s" % " ".join([str(a) for a in args]))

    def show(zelf):
        """ list of key,value pairs. """
        return ["%s=%s" % (a, b) for a, b in zelf.items()]

    ## ITERATORS

    def objects(zelf, *args, **kwargs):
        """ list of all object's. """
        if "time" in kwargs: desired_time = kwargs["time"]
        else: desired_time = ""
        objs = []
        path = zelf.get_root(kernel.cfg.workdir)
        for fnn in list_files(path, **kwargs):
            obj = Object().load(fnn, **kwargs)
            if args and args[0] not in obj: continue
            if desired_time and desired_time not in obj.get_timed(): continue
            if "deleted" in obj and obj.deleted: continue
            objs.append(obj)
        return objs

    def selected(zelf, *args, **kwargs):
        """ list of desired objects. """
        parsed = args[0]
        if kernel.cfg.debug: logging.warn("> parse %s" % parsed.pretty())
        if "exclude" in kwargs: exclude = kwargs["exclude"]
        else: exclude = None
        if exclude in parsed.args: parsed.args.remove(exclude)
        if "time" in parsed.wanted: kwargs["time"] = parsed.wanted.time
        objs = []
        for obj in zelf.objects(**kwargs):
            if not obj.check_wanted(parsed.wanted): continue
            if not "full" in kwargs and not obj.selector(parsed.args): continue
            if obj.check_notwanted(parsed.not_wanted): continue
            objs.append(obj)
        return objs

    ## SELECTOR

    def selector(zelf, *args, **kwargs):
        """ see if this objects has the desired attributes. """        
        wanted = args[0]
        if not wanted: return True
        go = False
        for w in wanted:
           if zelf.get(w, ""): go = True ; break
        return go


    ## TIME RELATED

    def first(zelf, *args, **kwargs):
        """ return the first object where the key and/or value matches. """
        key = args[0]
        try: value = args[1]
        except IndexError: value = None
        result = None
        for o in zelf.objects():
            if key not in o: continue
            if value and o[key] != value: continue
            if not result: result = o
            if o.get("timed", "") < result.get("timed", ""): result = o
        return result

    def last(zelf, *args, **kwargs):
        """ return the last object where the key and/or value matches. """
        key = args[0]
        try: value = args[1]
        except IndexError: value = None
        result = None
        for o in zelf.objects():
            if key not in o: continue
            if value and value != o[key]: continue
            if not result: result = o
            if a_time(short_date(o.timed)) > a_time(short_date(result.timed)): result = o
        return result

    ## WAITERS

    def ready(zelf):
        """ signal to ready state. """
        logging.debug("ready %s" % get_clsname(zelf))
        zelf._ready.set()

    def clear(zelf):
        """ clear the ready state. """
        logging.debug("clear %s" % get_clsname(zelf))
        zelf._ready.clear()

    def wait(zelf, sec=180.0):
        """ wait for ready state. """
        logging.debug("wait %s" % get_clsname(zelf))
        try: zelf._ready.wait(sec)
        except: pass

    ## HELPERS

    def to_json(zelf, *args, **kwargs):
        """ JSON string representation of this object. """
        return json.dumps(zelf.get_obj(*args, **kwargs), default=smooth, *args, **kwargs)

    def to_full(zelf, *args, **kwargs):
        """ full JSON dump of this object. """
        return json.dumps(zelf, default=smooth, indent=4)

    def make_signature(zelf, sig=None):
        """ signature of the data contained in this object. """
        if "data" in zelf: return str(hashlib.sha1(bytes(str(zelf.data), "utf-8")).hexdigest())
        raise NotSet

    def make_path(zelf, *args, **kwargs):
        """ create workdir if necessary. """
        path = zelf.get_path(*args, **kwargs) ; cdir(path) ; return path

    def pretty(zelf):
        """ nice formatted JSON string of this object. """
        return json.dumps(zelf.get_obj(), indent=2, default=pretty, sort_keys=True, ensure_ascii=True)

## TASKS

class Runner(threading.Thread):

    """ the working unit of CORELIB.  """

    nr = 0

    def __init__(zelf, *args, **kwargs):
        threading.Thread.__init__(zelf, None, zelf._loop, "Runner-%s" % str(zelf.nr), args, kwargs)
        zelf.setDaemon(True)
        zelf._queue = queue.Queue()
        zelf._outqueue = queue.Queue()
        zelf._ready = threading.Event() 
        zelf._state = Object()
        zelf._state.boot = time.time()
        zelf._state.output = time.time()
        zelf._state.input = time.time()
        zelf._status = Object()
        zelf._status.status = "starting"
        zelf._status.cmnd = "none"
        zelf.nr += 1

    def _loop(zelf):
        """ loop that executes the (function, object) pairs. """
        while zelf._status.status:
            zelf._status.status = "running"
            args, kwargs = zelf._queue.get()
            if not zelf._status.status: break
            func = args[0]
            try: obj = args[1]
            except: obj = Object()
            zelf._status.status = "exec"
            zelf._status.cmnd = get_funcname(func)
            zelf._state.input = time.time()
            func(*args[1:], **kwargs)
            if kernel.cfg.debug: logging.warn("> func %s" % str(obj.pretty()))
            obj.ready()
            time.sleep(0.01)
        _thread.interrupt_main()

    ## WAITERS

    def ready(zelf):
        """ signal to ready state. """
        logging.debug("ready %s" % get_clsname(zelf))
        zelf._ready.set()

    def clear(zelf):
        """ clear the ready state. """
        logging.debug("clear %s" % get_clsname(zelf))
        zelf._ready.clear()

    def wait(zelf, sec=180.0):
        """ wait for ready state. """
        logging.debug("wait %s" % get_clsname(zelf))
        try: zelf._ready.wait(sec)
        except: pass

    def exit(zelf, *args, **kwargs):
        """ stop the Runner. """
        zelf._state.status = ""

    def put(zelf, *args, **kwargs):
        """ put arguments/kwargs to the Runner. """
        zelf._queue.put_nowait((args, kwargs))
        return args[1]

    def get_name(zelf, *args, **kwargs): return "%s/%s" % (zelf.name, zelf._status.cmnd)

## DISPATCH

class Dispatcher(Object):

    """ the dispatcher delegates the workload to the Runners. Runner gets instantiated when needed. """

    def __init__(zelf, *args, **kwargs):
        Object.__init__(zelf, *args, **kwargs)
        zelf._runners = collections.deque()
        zelf._state = Object()
        zelf._state.boot = time.time()
        zelf._state.input = time.time()
        zelf._state.output = time.time()
        zelf._status.status = "starting"
        zelf._status.cmnd = ""
        zelf._max = 50

    def is_alive(zelf, *args, **kwargs):
        """ whether any Runners running. """
        for runner in zelf._runners:
           if runner.is_alive(): return True
        return False

    def register(zelf, *args, **kwargs):
        """ callback type with corresponding callback function. """
        o = Object()
        o.cbtype = args[0]
        o.func = args[1]
        logging.info("register %s" % o.how)
        if o.cbtype not in zelf: zelf[o.cbtype] = []
        zelf[o.cbtype].append(o)

    def denied(zelf, *args, **kwargs):
        for obj in zelf.objects("denied"):
            if args[0] == obj.get("denied", None): return True
        return False

    def allowed(zelf, *args, **kwargs):
        for obj in zelf.objects("allowed"):
            if args[0] == obj.get("allowed", None): return True
        return False
 
    def dispatch(zelf, *args, **kwargs):
        """ dispatch an event onto a Runner. """
        event = args[0]
        zelf.execute(event.get_cmnd(), event)

    def execute(zelf, *args, **kwargs):
        """ execute a command, event pair. """
        cmnd = args[0]
        event = args[1]
        if cmnd not in zelf: return
        if "origin" in event:
            if zelf.allowed(event.origin): 
                logging.warn("! allow %s" % event.origin)
            elif zelf.denied(event.origin):
                logging.warn("! deny %s" % event.origin)
                return
        logging.info("> exec %s" % cmnd)
        for functor in zelf[cmnd]:
            if kernel.cfg.shell: kernel.workers.put(functor.func, event)
            else: functor.func(event)

    def run_func(zelf, *args, **kwargs):
        """ execute command/event pair. """
        cmnd = args[0]
        event = args[1]
        if cmnd not in zelf: return
        for functor in zelf[cmnd]: functor.func(event)

    def exit(zelf, name=None):
        """ stop the Runners. """
        for runner in zelf._runners:
            if name and name not in runner.name: continue
            runner.join(0.1)

    def put(zelf, *args, **kwargs):
        """ put load to the Runner. """
        zelf.cleanup()
        target = zelf.make_new()
        return target.put(*args, **kwargs)

    def make_new(zelf, *args, **kwargs):
        """ create a Runner. """
        if len(zelf._runners) < zelf._max:
            runner = Runner(*args, **kwargs)
            zelf._runners.append(runner)
            runner.start()
        else: runner = random.choice(zelf._runners)
        return runner

    def cleanup(zelf, dojoin=False):
        """ remove idle Runners. """
        todo = []
        for runner in zelf._runners:
            if not runner._status.status or not runner._queue.qsize(): todo.append(runner)
        for runner in todo: runner.exit() ; zelf._runners.remove(runner)

## PLUGINS

class Plugins(Object):

    """ object to register plugins with. """

    def __init__(zelf, *args, **kwargs):
        Object.__init__(zelf, *args, **kwargs)
        zelf.plugins = []

    def get_names(zelf, plugsdir):
        """ plugnames from the plugin directory. """
        return [x[:-3] for x in os.listdir(plugsdir) if x.endswith(".py")]

    def load_plugs(zelf, name):
        """ load plugins from the plugin directory. """
        logging.warn("load %s" % name)
        plugname = name.split(".")[-1]
        mod = __import__(name)
        path, fn = os.path.split(mod.__file__)
        path = path + os.sep + plugname 
        for plugname in zelf.get_names(path):
            if "__" in plugname: continue
            try:
                zelf[plugname] = zelf.load_mod(plugname, path, force=True)
                if plugname not in zelf.plugins: zelf.plugins.append(plugname)
            except: error() ; continue

    load_package = load_plugs

    def load_mod(zelf, plugname, pdir="", force=True):
        """ load module. """
        if plugname in zelf and not force:
            if not force: return zelf[plugname]
            zelf[plugname] = imp.reload(zelf[plugname])
        else:
            if not pdir: pdir = j(zelf.root, "plugs")
            search = imp.find_module(plugname, [pdir,])
            zelf[plugname] = imp.load_module(plugname, *search)
        return zelf[plugname]

    def plug_init(zelf, *args, **kwargs):
        """ initialize plugin. """
        plugname = args[0]
        if "init" in dir(zelf[plugname]):
            logging.warn("init %s" % plugname)
            zelf[plugname].init(*args, **kwargs)
            
    def plug_shutdown(zelf, plugname): 
        """ call the shutdown() function, if provided. """
        if plugname not in zelf: return
        if "shutdown" in dir(zelf[plugname]): zelf[plugname].shutdown()

    def exit(zelf, *args, **kwargs):
        """ exit all plugins. """
        for plugname in zelf: zelf.plug_shutdown(plugname)

    def reload(zelf, plugname, force=False):
        """ reload a plugin. """
        zelf.unload(plugname)
        mod = zelf.load_mod(plugname, force)
        return mod

    def init(zelf, *args, **kwargs):
        """ init all plugins or a specific one if argument is given. """
        try: targets = args[0].split(",")
        except ValueError: targets = [args[0], ]
        except IndexError: targets = []
        launched = []
        for plugname in zelf.plugins:
            if "all" not in targets and plugname not in targets: continue 
            zelf.plug_init(plugname, *args, **kwargs)
        zelf.ready()

## BOOT

def boot(*args, **kwargs):
    """ get the program properly initialized. """
    global kernel
    global workdir
    if args: cfg = args[0]
    else: cfg = Object()
    cfg.loglevel = cfg.get("loglevel", "error")
    core.log.init(cfg.loglevel)
    if not cfg.workdir: cfg.workdir = homedir + os.sep + "core.data"
    if cfg.get("eggs", None): cfg.workdir = os.path.expanduser(j(os.getcwd(), "data"))
    cdir(cfg.workdir)
    kernel.cfg.update(cfg)
    if kernel.cfg.loglevel != "error":
        for line in kernel.cfg.show(): logging.warn(line)
    if kernel.cfg.loglevel != "error":
        print("")
        print("%sP L U G I N S%s" % (BLUE, ENDC))
        print("")
    kernel.plugs.load_plugs("core.plugs")
    kernel.plugs.init(cfg.get("init", ""))
    kernel.plugs.wait()
    kernel.ready()
    return kernel

## SHUTDOWN

def shutdown():
    """ close the program in a appropiate manner. """
    global kernel
    print("")
    readline.write_history_file(histfile)
    for obj in kernel.services:
        if obj == None: continue
        try: obj.exit()
        except AttributeError: pass
        except: error()
    kernel.cmnds.exit()
    kernel.plugs.exit()
    if kernel.cfg.shell: 
        logging.warn("exit")
    os._exit(0)

## KERNEL

kernel = Object()
kernel.cmnds = Dispatcher()
kernel.workers = Dispatcher()
kernel.plugs = Plugins()
kernel.services = Object()
kernel.cfg = Object()
kernel.fleet = []
